--[[
	Description: DataMiner is a WoW-Addon to collect loot data while doing herbalism or mining professions.
	Copyright (C) 2014, 2015 Linuz@Proudmore.Eu
    This program is free software; you can redistribute it and/or modify it under the terms of the
	GNU General Public License as published by the Free Software Foundation; either version 2 of the License,
	or (at your option) any later version.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program;
	if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--]]

-- Library references
local L = LibStub("AceLocale-3.0"):GetLocale("DataMiner", true)

local Addon = "DataMiner"
local titan = IsAddOnLoaded( "Titan" )
local mapId								-- the current zone of the character
local search							-- searchfield for viewmode 8,17
local viewmode = 7						-- 00000 Bitfield viewmode
										-- ║║║║╚═ aMapId
										-- ║║║╚══ aSeg
										-- ║║╚═══ aDay
										-- ║╚════ aName
										-- ╚═════ Cumulative list

function DM_Create_Tip(dm_base)
	local iDay = date("%A")
	local iSeg = TimeSegment(8)
	local maxlootrate = 0
	local iMapId
	
	for idx, line in ipairs(dm_base) do
		if ((line.aDay == iDay) and (line.aSeg == iSeg) and (line.aLoot > 10)) then
			if (maxlootrate < (line.aCount / line.aLoot)) then
				maxlootrate = line.aCount / line.aLoot
				iMapId = line.aMapId
			end
		end
	end
	if (maxlootrate > 1.5) then
		local res
		if (dm_base == dm_herbs) then res = L["herbs"] end
		if (dm_base == dm_ore) then res = L["ore"] end
		print(L["Go_to"].." "..GetMapNameByID(iMapId).." "..L["for"].." "..res.." "..
			  L["because_lootrate_there_is"]..string.format(" %.2f ", maxlootrate)..L["now_"])
	end
end

local function SetViewmode( vmode, s )
	if ( vmode == 7 or vmode == 0) then
		viewmode = 7 search = nil
	elseif( vmode == 8 ) then
		viewmode = 8 search = s
	elseif( vmode == 17 ) then
		viewmode = 17 search = s
	end
	if (titan) then Titan_SetTitle( vmode, s )
	else LDB_SetTitle( vmode, s ) end
end

function ToggleDataMining(self)
	local hasherbalism, hasmining = HasProfession(PROF_HERBALISM, false), HasProfession(PROF_MINING, false)
	if (not hasherbalism and not hasmining) then return end
	do
		local states = 1
		if (hasherbalism) then states = inc(states) end
		if (hasmining) then states = inc(states) end
		collect = ToggleX(states, collect)
	end
	
	if (collect) then
		SetViewmode(7)		--when we switch datamining, we reset the viewmode to default
		self:RegisterEvent("LOOT_OPENED")
		self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		--mapid's are unique numeric identifiers for zone maps.
		--They can be obtained by calling the GetCurrentMapAreaID.
		--These value can be passed to the GetMapNameByID API to identify zones.
		SetMapToCurrentZone()
		mapId = GetCurrentMapAreaID()
		DM_Log_Msg(102)
		if (hasherbalism) then DM_Create_Tip(dm_herbs) end
		if (hasmining) then DM_Create_Tip(dm_ore) end
	else
		SetViewmode(0)		--when we switch datamining, we reset the viewmode to default
		self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		self:UnregisterEvent("LOOT_OPENED")
		DM_Log_Msg(103)
	end
end

function DM_GetTooltipText()
	local function GetIdx(buff, id) -- for viewmode 17
		for idx, line in ipairs(buff) do
			if (line.aItemID == id) then
				return idx
			end
		end
		return 0
	end
		
	-- Which table should we use
	local dm_base
	if ( not collect ) then
		return L["Currently_mining_data_is"].." "..GetHexColoredText(L["off"], COL_GREEN2)
	elseif ( collect == 1 ) then
		if (HasProfession(PROF_HERBALISM, false)) then dm_base = dm_herbs
		elseif (HasProfession(PROF_MINING, false)) then dm_base = dm_ore end
	elseif ( collect == 2 )	then
		if (HasProfession(PROF_MINING, false)) then dm_base = dm_ore end
	end
	if (dm_base == nil) then
		return L["This_character_has_no_supported_profession"]
	end
	
	local dm_buff = {}
	local iDay = date("%A")
	local iSeg = TimeSegment(8)
	local iMapId = mapId
	local out = ""
	local spacer = "-------------------------------------------------------\n"
	local header
	local maxrows = 30      -- We don't have a scrollbar, so we limit the rows of output
	
	if (viewmode == 7) then -- view name count loots rate
		header = spacer
				 ..GetHexColoredText(L["Item"],COL_YELLOW).."\t"
				 ..GetHexColoredText(L["Pieces"].."  ",COL_GREEN)
				 ..GetHexColoredText(L["Places"].."  ",COL_WHITE)
				 ..GetHexColoredText(L["Rate"],COL_ORANGE)
				 .."\n"..spacer
		for idx, line in ipairs(dm_base) do
			if ((line.aMapId == iMapId) and (line.aSeg == iSeg) and (line.aDay == iDay)) then
				table.insert (dm_buff, dm_base[idx])
			end
		end
		table.sort(dm_buff, function(a, b) return (a.aCount/a.aLoot) > (b.aCount/b.aLoot) end)
		for idx, line in ipairs(dm_buff) do
			out = out..GetHexColoredText(GetItemInfo(dm_buff[idx].aItemID), COL_YELLOW).."\t"..
		               GetHexColoredText(dm_buff[idx].aCount, COL_GREEN).."  "..
					   GetHexColoredText(format("%6.d",dm_buff[idx].aLoot), COL_WHITE).."  "..
					   GetHexColoredText(format("%.2f",dm_buff[idx].aCount / dm_buff[idx].aLoot), COL_ORANGE).."\n"
			-- Limit the output to maxrows lines
			if ( idx == maxrows ) then break end
		end
	elseif (viewmode == 8) then -- item view
		spacer = "-----------------------------------------------------------------\n" --we need more space here
		header = spacer
				..GetHexColoredText(L["A_Time"].."  ",COL_GREEN)
				..GetHexColoredText(L["Weekday"],COL_YELLOW).."\t"
				..GetHexColoredText(L["Zone"].."  ",COL_WHITE)
				..GetHexColoredText(L["Rate"],COL_ORANGE)
				.."\n"..spacer
		for idx, line in ipairs(dm_base) do
			-- If not in cash GetItemInfo returns nil, in this
			-- case we have no chance to get the name!!
			if GetItemInfo( dm_base[idx].aItemID ) then
				if string.lower( GetItemInfo(dm_base[idx].aItemID)) == search then
					table.insert( dm_buff, dm_base[idx] )
				end
			else --[[return header..L["Try_again_later_item_is_not_cashed"]--]] end
			-- todo: If we uncomment the above all items must be in cash!!!
		end
		
		table.sort(dm_buff, function(a, b) return (a.aCount/a.aLoot) > (b.aCount/b.aLoot) end)
		for idx, line in ipairs(dm_buff) do
			out = out..GetHexColoredText(dm_buff[idx].aSeg, COL_GREEN).."  "..
					   GetHexColoredText(dm_buff[idx].aDay, COL_YELLOW).."\t"..
					   GetHexColoredText(GetMapNameByID(dm_buff[idx].aMapId), COL_WHITE).."  "..
					   GetHexColoredText(format("%.2f",dm_buff[idx].aCount / dm_buff[idx].aLoot), COL_ORANGE).."\n"
			-- Limit the output to maxrows lines
			if ( idx == maxrows ) then break end		   
		end
	elseif (viewmode == 17) then  -- Cumulative Zone View
		local idx
		header = spacer
				 ..GetHexColoredText(L["Item"],COL_YELLOW).."\t"
				 ..GetHexColoredText(L["Pieces"].."  ",COL_GREEN)
				 ..GetHexColoredText(L["Places"].."  ",COL_WHITE)
				 ..GetHexColoredText(L["Rate"],COL_ORANGE)
				 .."\n"..spacer
		-- We copy/alter the needed lines to dm_buff
		for i, line in ipairs(dm_base) do
			-- If not in cash GetItemInfo returns nil, in this
			-- case we have no chance to get the name!!
			if (GetItemInfo(line.aItemID)) then
				if (string.lower(GetMapNameByID(line.aMapId)) == search) then
					idx = GetIdx( dm_buff, line.aItemID )                 -- Look if this item is already there
					if (idx>0) then   									  -- Already there
						-- We want to cumulate the items
						dm_buff[idx].aCount = dm_buff[idx].aCount + line.aCount
						dm_buff[idx].aLoot = dm_buff[idx].aLoot + line.aLoot
					else                                                  -- This is a new entry
						table.insert (dm_buff, line)
					end
				end
			end
		end
		table.sort(dm_buff, function(a, b) return (a.aCount/a.aLoot) > (b.aCount/b.aLoot) end)
		for idx, line in ipairs(dm_buff) do
			out = out..GetHexColoredText(GetItemInfo(dm_buff[idx].aItemID), COL_YELLOW).."\t"..
		               GetHexColoredText(dm_buff[idx].aCount, COL_GREEN).."  "..
					   GetHexColoredText(format("%6.d",dm_buff[idx].aLoot), COL_WHITE).."  "..
					   GetHexColoredText(format("%.2f",dm_buff[idx].aCount / dm_buff[idx].aLoot), COL_ORANGE).."\n"
			-- Limit the output to maxrows lines
			if ( idx == maxrows ) then break end
		end
	end
			
	if (out == "") then
		out = L["No_corresponding_data_found"].."\n"
	end
	return header..out
end

-------------
-- Loging --
-------------

-- Add a new line to log
-- If the maximum capacity is reached, then delete old entrys
function DM_Log_Add(iLog)
	if (not logsize) then logsize = 50 end
	if ( #dm_log >= logsize ) then
		repeat
			table.remove(dm_log, 1)         		-- del oldest msg until size is valid
		until #dm_log <= logsize
	end
	table.insert (dm_log, {aLog = iLog})
end

-- Encapsulate all log messages
-- so we have a single place to hold them
function DM_Log_Msg(num, ext)
	if (not logg) then return end	-- we leaving all untouched if log is off
	if (num == 99) then
		DM_Log_Add(Addon.." v"..GetVersion().." "..L["loaded"])
		print(Addon.." v"..GetVersion().." "..L["loaded"])
	elseif (num == 100) then
		DM_Log_Add(L["Loging_is_now"].." "..L["off"])
	elseif (num == 101) then
		DM_Log_Add(L["Loging_is_now"].." "..L["on"])
	elseif (num == 102) then
		DM_Log_Add(L["DataMining_is_now"].." "..L["on"])
	elseif (num == 103) then
		DM_Log_Add(L["DataMining_is_now"].." "..L["off"])
	elseif (num == 104) then
		DM_Log_Add(L["Updating"]..ext)
	elseif (num == 105) then
		DM_Log_Add(L["Adding"]..ext)
	elseif (num == 106) then
		DM_Log_Add(L["Canceling"]..ext)
	elseif (num == 107) then
		DM_Log_Add(L["Removing_from_table_at_index"]..ext)
	elseif (num == 108) then
		DM_Log_Add(L["Clearing_log"])
	elseif (num == 109) then
		DM_Log_Add(L["Entering"]..ext)
	else
		-- this should never reached
		print("Error: No Msg")
	end
end

-- Enable/Disable the loging feature
function DM_Toggle_Log()
	if (logg) then
		logg = false
		DM_Log_Msg(100)
	else
		logg = true
		DM_Log_Msg(101)
	end
end

-- print log to chat
function DM_Log_Print()
	if ( #dm_log == 0 ) then return end
	print(Addon.." log:")
	for idx, line in ipairs(dm_log) do
		print(idx.." "..line.aLog)
	end
end

-------------------
-- Chat Commands --
-------------------

-- Print the herbs or ore table to chat
function DM_Table_Print( dm_base )
	if ( #dm_base == 0) then return end
	if (dm_base == dm_herbs) then
		print(Addon.." "..L["Herbalism_Statistics"]..":")
	elseif (dm_base == dm_ore) then
		print(Addon.." "..L["Mining_Statistics"]..":")
	end
	for idx, line in ipairs(dm_base) do
		print(idx, line.aDay, line.aSeg, line.aMapId, line.aItemID, " x"..line.aCount, line.aLoot)
	end
end

function DM_RunCommand( key )
	if (not key) then
		local editbox = ChatEdit_ChooseBoxForSend(DEFAULT_CHAT_FRAME)
		ChatEdit_ActivateChat( editbox )
		editbox:SetText("/join "..Addon)
		ChatEdit_OnEnterPressed(editbox)
		return
	end
	for idx, line in ipairs(dm_sc) do
		if (line.aKey == string.upper(key)) then  -- we only execute the first command
			local editbox = ChatEdit_ChooseBoxForSend(DEFAULT_CHAT_FRAME)
			ChatEdit_ActivateChat( editbox )
			editbox:SetText(line.aCommand)
			ChatEdit_OnEnterPressed(editbox)
			return
		end
	end
end

-- Builds a parameter list from chat input
-- Parameter: str     e.g /dm view item "my cool item"
-- Return: argc, args[argc]
local function DM_Slash_Parser(str)
	local param, first, second
	local q_open = false
	local argc = 0
	local args = {}
		
	if ( str == "" ) then return argc, args end
	for param in string.gmatch(str, '[%w\128-\244\"\':]+') do -- we need %w for /dm table remove <#>
		first = string.find(param, "\"")                      -- \128-\244 to support umlauts
		second = string.find(param, "\"", 2)				  -- : for shortcuts
		if (first == second) then first = false end
		if (first and second) then
			param = string.sub(param, 2, second-first)        -- param is quoted
			argc = inc(argc); args[argc] = param
		elseif (first and (not second)) then                  -- param is quoted beginning part
			param = string.sub(param, 2)
			argc = inc(argc); args[argc] = param
			q_open = true                                     -- q_open: needed for quoted 3-part or more param
		elseif ((not first) and second) then                  -- param is quoted ending part
			param = string.sub(param, 1, dec(second))
			args[argc] = args[argc].." "..param
			q_open = false
		else
			if (q_open) then                                 -- param is quoted middle part
				args[argc] = args[argc].." "..param
			else
				argc = inc(argc); args[argc] = param         -- param is not quoted
			end
		end
	end
	return argc, args
end

function DM_SlashHandler(msg)
	local args, argc = {}
	argc, args = DM_Slash_Parser(msg)
	
	if ((args[1] == "help") or (argc == 0)) then
		print(L["Available_commands"]..":")
		print("/dm "..L["or"].." /dm help                     - "..L["This_message"])
		print("/dm toggle          ".."                      - "..L["Toggle_datamining"])
		print("/dm table herbs|ore ".."                   - "..L["Show_table"])
		print("/dm table herbs|ore remove <n>".."  - "..L["Removes_index_n_from_table"])
		print("/dm time            ".."                      - "..L["Show_current_time_segment"])
		print("/dm log             ".."                       - "..L["Show_log"])
		print("/dm log delete       ".."                    - "..L["Delete_log"])
		print("/dm log toggle      ".."                    - "..L["Toggle_loging"])
		print("/dm view default    ".."                   - "..L["Default_view"])
		print("/dm view item|zone <name>".."       - "..L["Item_view"])
	elseif (isStart(args[1], ":" ) and isEnd(args[1], ":") and (argc == 1)) then
		DM_RunCommand(args[1]:match("%w+"))       -- this is a shortcut to execute
	elseif (args[1] == "view") then
		if (args[2] == "default") then
			SetViewmode( 7 )                      --views current zone,day,time
		elseif ((args[2] == "item") and (argc == 3)) then
			SetViewmode( 8, args[3] )             -- view name
		elseif ((args[2] == "zone") and (argc == 3)) then
			SetViewmode( 17, args[3] )            -- views cumulative zone
		end
	elseif (args[1] == "toggle") then
		ToggleDataMining(self)
	elseif (args[1] == "time") then
		print (L["Current_time_segment_is"].." "..TimeSegment(8))
	elseif (args[1] == "table") then
		if (args[2] == "herbs")  then
			if (argc == 2) then
				DM_Table_Print( dm_herbs )
			elseif ((args[3] == "remove") and (argc == 4)) then
				table.remove(dm_herbs, tonumber(args[4]))
				DM_Log_Add(107, " "..args[4])
			end	
		elseif (args[2] == "ore") then
			if (argc == 2) then
				DM_Table_Print( dm_ore )
			elseif ((args[3] == "remove") and (argc == 4)) then
				table.remove(dm_ore, tonumber(args[4]))
				DM_Log_Msg(107, " "..args[4])
			end	
		end
	elseif (args[1] == "log") then
		if ((argc == 1) or ((argc == 2) and (args[2] == "show"))) then
			DM_Log_Print()
		elseif (args[2] == "delete") then
			Table_reset(dm_log)
			DM_Log_Msg(108)
		elseif (args[2] == "toggle") then
			DM_Toggle_Log()
		end
	end
end

------------
-- Events --
------------

-- Collect wanted data and build or update tables
function DM_Data_Mining()
	local function getIndex(dm_base, iId, iMapId, iDay, iSeg)
		for i, line in ipairs(dm_base) do
			if ((line.aItemID == iId) and
				(line.aMapId == iMapId) and
				(line.aDay == iDay) and
				(line.aSeg == iSeg)) then
				return i
			end
		end
		return 0
	end
	
	local idx
	local itemIcon, itemName, itemCount, itemQuality
	local itemLink, itemId
	local iDay = date("%A")
	local iSeg = TimeSegment(8)
	local dm_base
	
	-- Which table should we use
	-- collect == 2 is the case, when we have both professions,
	-- but herbalism is prof-#1 in any case
	if ( collect == 1 ) then
		if (HasProfession(PROF_HERBALISM, false)) then dm_base = dm_herbs
		elseif (HasProfession(PROF_MINING, false)) then dm_base = dm_ore end
	elseif ( collect == 2 )	then
		if (HasProfession(PROF_MINING, false)) then dm_base = dm_ore end
	end
	
	for i=1, GetNumLootItems(), 1 do
		itemLink = GetLootSlotLink(i)
		itemId = itemLink and tonumber(strmatch(itemLink, "item:(%d+)"))
		itemIcon, itemName, itemCount, itemQuality = GetLootSlotInfo(i)      -- We only use name, count
		
		if ((IsHerbOre(itemId, BAG_HERBALISM)) and (dm_base == dm_herbs) or
			(IsHerbOre(itemId, BAG_MINING)) and (dm_base == dm_ore)) then
			idx = getIndex(dm_base, itemId, mapId, iDay, iSeg)     			 --Look if this item is already there
			if ( idx > 0 ) then                                         		 --Already there
				PlaySoundFile("Interface\\AddOns\\TitanDataMiner\\sounds\\update.mp3")
				dm_base[idx].aCount = dm_base[idx].aCount + itemCount
				dm_base[idx].aLoot = inc(dm_base[idx].aLoot)
				DM_Log_Msg(104, " "..itemName.." ID = "..itemId.." Zone = "..GetMapNameByID(mapId))
			else                                                    		 --This is a new entry
				PlaySoundFile("Interface\\AddOns\\TitanDataMiner\\sounds\\add.mp3")
				table.insert (dm_base, {aItemID = itemId, aMapId = mapId, aDay = iDay, aSeg = iSeg, aCount = itemCount, aLoot = 1})
				DM_Log_Msg(105, " "..itemName.." ID = "..itemId.." Zone = "..GetMapNameByID(mapId))
			end
		else
			-- We don't want to spam the chat
			--PlaySoundFile("Interface\\AddOns\\TitanDataMiner\\sounds\\cancel.mp3")
			--DM_Log_Msg(106, " "..itemName.." ID = "..itemId.." Zone = "..GetMapNameByID(mapId))
		end
	end
end

function DM_OnEvent(ev)
	if (ev == "LOOT_OPENED") then
		DM_Data_Mining()
	elseif (ev == "ZONE_CHANGED_NEW_AREA") then
		SetMapToCurrentZone()
		mapId = GetCurrentMapAreaID()
		DM_Log_Msg(109, " "..GetMapNameByID(mapId))
	elseif (ev == "PLAYER_ENTERING_WORLD") then
		-- Enable Slash Commands
		SlashCmdList["DataMiner"] = function(msg) DM_SlashHandler(string.lower(msg)) end
		SLASH_DataMiner1 = "/dm"
		SLASH_DataMiner2 = "/dataminer"
		-- clear the chat
		SlashCmdList["CLR"] = function() local s = "" for i = 1, 512 do s = s..'\r' end print(s) end
		SLASH_CLR1 = "/clr"
		SLASH_CLR2 = "/clear"
		
		DM_Log_Msg(99)
		SetupConfig()
		DM_RunCommand()
	else
		-- this should never reached
		print("Error: Unknown Event "..ev)
	end
end

-- Called from xml code on mouse event
function DM_OnClick(self, mousebutton)
	if (titan) then -- native
		if (mousebutton == "LeftButton") then
			ToggleDataMining(self);
		elseif (mousebutton == "RightButton") then
			TitanPanelDMButton_OnClick(self, mousebutton)
		end
	else  -- LDB should never reached, as of mouseevents handled through ldb struct
		print(mousebutton.."pressed")
	end
end

function DM_OnLoad(self)
	if (titan) then
		TitanPanelDMButton_OnLoad(self)
		TitanPanelButton_OnLoad(self)
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
	else
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		dm_ldb.Button_OnLoad(self)
	end
end
